import torch
from torch.utils.data.dataset import Dataset
from anc.data.parquet_dataset import ParquetConcateDataset
from anc.data.utils import DataNotReady
from torch.utils.data import IterableDataset
import inspect
import itertools
import logging


class AncDataset(IterableDataset):
    r'''
    A class defines the pipeline of data loading and data transform

    Args:
        filepaths (list of str): files from which to load the data.
        processor (Processor): object that defines the data transform and batch transform logic.
        sampler: object that control the data loader order
        dataset_type (str, optional): the type of dataset that the paths represents (default ``parquet``).
        ds_args (dict, optional): dataset arguments for data loading, such as the columns to load 
            for parquet files, etc (default ``{}``).
        repeat (bool, optional): if set to True, the data will be loaded repeatedly
    '''
    def __init__(self, filepaths, processor, sampler=None, dataset_type="parquet", ds_args={}, repeat=False):
        self.filepaths = filepaths
        self.dataset_type = dataset_type
        self.ds_args = ds_args
        # TODO: support more types of dataset
        self.ds = ParquetConcateDataset(filepaths, columns=ds_args.get("columns"))
        self.processor = processor
        self.sampler = sampler
    
    def set_sampler(self, sampler):
        self.sampler = sampler
    
    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        wid = worker_info.id if worker_info is not None else 0
        num_workers = worker_info.num_workers if worker_info is not None else 1
        logging.debug(f"iter information: len(sampler) - {len(self.sampler)}, worker id - {wid}, num_workers - {num_workers}")
        # use islice to get the iterator of sampler only for current worker
        sampler_iter = itertools.islice(iter(self.sampler), wid, len(self.sampler), num_workers)
        yield from self.get_generator(sampler_iter)

    def get_generator(self, sampler_iter):
        while True:
            indices = next(sampler_iter)
            item_generator = self._getitem(indices)
            is_last = False
            while True:
                try:
                    wid, next_data, is_last = next(item_generator)
                    yield (wid, [next_data], False)
                except StopIteration:
                    if is_last:
                        yield (wid, DataNotReady(), True)
                    break
            if is_last:
                break

    def _getitem(self, idxs):
        worker_info = torch.utils.data.get_worker_info()
        wid = worker_info.id if worker_info is not None else 0
        idxs, is_last_batch = idxs
        if isinstance(idxs, int):
            idxs = [idxs]
        if len(idxs) == 0:
            yield (wid, DataNotReady(), is_last_batch)
            return
        raw_data = [self.ds[idx] for idx in idxs]
        if self.processor.transform:
            processed_data = []
            for data_idx, data in enumerate(raw_data):
                is_last_sample = is_last_batch and data_idx == len(raw_data) - 1
                processed = self.processor.transform(data, is_last_sample)
                if processed is None:
                    continue
                if isinstance(processed, list):
                    processed_data += processed
                else:
                    # processed_data is a generator
                    if self.processor.batch_transform:
                        # batch_processed is a generator too
                        batch_processed = self.processor.batch_transform(processed, is_last_sample)
                        if batch_processed is not None:
                            for batch in batch_processed:
                                yield (wid, batch, is_last_sample)         
        else:
            processed_data = raw_data
        if processed_data and self.processor.batch_transform:
            batch_processed = self.processor.batch_transform(processed_data, is_last_batch)
            if batch_processed is None:
                batch_processed = DataNotReady()
            else:
                assert isinstance(batch_processed, list)
        else:
            batch_processed = []
        for batch in batch_processed:
            yield (wid, batch, is_last_batch)

    def __getstate__(self):
        state = {}
        state['filepaths'] = self.filepaths
        state['dataset_type'] = self.dataset_type
        state['ds_args'] = self.ds_args
        state['processor'] = self.processor.__getstate__()

    def __setstate__(self, state):
        if 'processor' in state:
            processor_state = state.pop('processor')
            self.processor.__setstate__(state['processor'])
        self.__dict__.update(state)

    def __len__(self):
        return len(self.ds)
    
    def get_sub_lengths(self):
        if self.dataset_type == 'parquet':
            level = "row_group"
        else:
            level = "file"
        return self.ds.get_sub_lengths(level)


if __name__ == "__main__":
    from anc.data.anc_processor import AncProcessor
    from anc.data.anc_sampler import AncSampler
    folder = "/mnt/personal/parquet_demo_data"
    fnames = [
        "01_0001.parquet",
        "02_0001.parquet",
        "03_0001.parquet",
        "04_0001.parquet",
        "05_0001.parquet",
        "06_0001.parquet",
        "07_0001.parquet"
    ]
    files = [f"{folder}/{fname}" for fname in fnames]
    ds = AncDataset(files, AncProcessor())
    sampler = AncSampler(ds, 1)
    ds.set_sampler(sampler)
    assert sum([sum(i) for i in ds.get_sub_lengths()]) == len(ds)
    ds_iter = iter(ds)
    for i in range(10):
        wid, data, is_last_batch = next(ds_iter)
        print(data[0][0]['filepath'])




import click
import json
import sys
from requests.exceptions import RequestException

from anc.cli.util import click_group
from anc.api.connection import Connection
from anc.conf.remote import remote_server
from .dataset_operator import DatasetOperator


@click_group()
def ds():
    pass


@ds.command()
@click.argument(
    "source_path",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True
    )
)
@click.option("--version", "-v", type=str, help="Dataset version you want to register", required=True)
@click.option("--message", "-m", type=str, help="Note of the dataset")
@click.pass_context
def add(ctx, source_path, version, message):
    """command like: anc ds add abc.txt -v 0.1"""
    op = DatasetOperator()
    op.remote_add(version, source_path, message)

@ds.command()
@click.option("--name", "-n", help="Name of the datasets in remote",)
def list(name):
    """command like: anc ds list"""
    op = DatasetOperator()
    op.list_dataset(name)


@ds.command()
#@click.option("--name", "-n", help="Name of the datasets in remote", required=True)
@click.argument(
    "name"
)
@click.option("--version", "-v", help="Version of the dataset")
@click.option("--dest", "-d", help="Destination path you want to creat the dataset",default=".", show_default=True)
@click.option("--cache_policy", "-c", help="If input is `no` which means no cache used, the dataset will be a completely copy")
@click.pass_context
def get(ctx, name, version, dest, cache_policy):
    """command like: anc ds get abc.txt -v 0.1"""
    op = DatasetOperator()
    op.download_dataset(name, version, dest, cache_policy)


@ds.group()
def queue():
    """Commands for operation task queue"""
    pass

@queue.command()
def status():
    """command like: anc ds queue status"""
    try:
        conn = Connection(url=remote_server)
        response = conn.get("/queue_status")

        if 200 <= response.status_code < 300:
            status_data = response.json()
            print("Queue Status:")
            print(json.dumps(status_data, indent=2))
        else:
            #print(f"Error: Server responded with status code {response.status_code}")
            print(f"Sorry, {response.text}")
    except RequestException as e:
        print(f"Sorry, you can't get queue info out of clusters, please use it in a notebook")
    except json.JSONDecodeError:
        print("Sorry: received invalid JSON response from server")
    except KeyboardInterrupt:
        print(f"Operation interrupted.")
        sys.exit(0)
    except Exception as e:
        print(f"Sorry, your command run failed, you can try again or reach out infra team")

@ds.group()
def task():
    """Commands for task operations"""
    pass

@task.command()
@click.argument("task_id", type=int, required=True)
def status(task_id):
    """command like: anc ds task status 201"""
    try:
        conn = Connection(url=remote_server)
        response = conn.get(f"/task_status/{task_id}")

        if 200 <= response.status_code < 300:
            status_data = response.json()
            print(f"Task Status for ID {status_data['task_id']}")
            print(json.dumps(status_data, indent=2))
        else:
            print(f"Error: Server responded with status code {response.status_code}")
            print(f"Sorry, {response.text}")
    except RequestException as e:
        print(f"Sorry, you can't run this command out of clusters, please use it in a notebook")
    except json.JSONDecodeError:
        print("Sorry: received invalid JSON response from server")
    except KeyboardInterrupt:
        print(f"Operation interrupted.")
        sys.exit(0)
    except Exception as e:
        print(f"Sorry, your command run failed, you can try again or reach out infra team")


@task.command()
@click.argument("task_id", type=int)
@click.option("--priority", "-p", type=int, required=True, help="New priority value for the task")
def boost(task_id, priority):
    """Set a new priority for a task"""
    try:
        conn = Connection(url=remote_server)
        data = {"new_priority": priority}
        response = conn.post(f"/task/{task_id}/increase_priority", json=data)

        if response.status_code == 200:
            result = response.json()
            print(result["message"])
        elif response.status_code == 400:
            error = response.json()
            print(f"Sorry: {error['error']}, maybe your task is not in pending state")
        elif response.status_code == 404:
            error = response.json()
            print(f"Sorry: {error['error']}, maybe your task is not in pending state")
        else:
            print(f"Sorry: Server responded with status code {response.status_code}")
            print(f"Response: {response.text}")
    except RequestException as e:
        print(f"Sorry, you can't get queue info out of clusters, please use it in a notebook")
    except json.JSONDecodeError:
        print("Sorry: received invalid JSON response from server")
    except KeyboardInterrupt:
        print(f"Operation interrupted.")
        sys.exit(0)
    except Exception as e:
        print(f"Sorry, your command run failed, you can try again or reach out infra team")


def add_command(cli_group):
    cli_group.add_command(ds)



import numpy as np
from functools import partial
from nemo.collections.nlp.data.language_modeling.megatron.indexed_dataset import IndexedCachedDataset
from nemo.collections.nlp.data.language_modeling.megatron.indexed_dataset import make_dataset as nemo_make_dataset
from nemo.collections.nlp.data.language_modeling.megatron.gpt_dataset import GPTDataset as NeMoGPTDataset
from nemo.collections.nlp.data.language_modeling.megatron.gpt_dataset import _build_doc_idx
from nemo.collections.nlp.data.language_modeling.megatron.gpt_dataset import build_dataset as nemo_build_dataset


def make_dataset(path, impl, skip_warmup=False, impl_kwargs={}, delay_data_mmap=False):
    if impl == 'sequential_cached':
        return SequentialIndexedCachedDataset(path)
    return nemo_make_dataset(path, impl, skip_warmup, impl_kwargs, delay_data_mmap)


def build_dataset(cfg, trainer, data_prefix, data_impl, num_samples, seq_length, seed, skip_warmup, tokenizer, name):
    import nemo.collections.nlp.data.language_modeling.megatron.gpt_dataset as nemo_gpt_dataset
    nemo_gpt_dataset.make_indexed_dataset = make_dataset
    return nemo_build_dataset(cfg, trainer, data_prefix, data_impl, num_samples, seq_length, seed, skip_warmup, tokenizer, name)


def _customized_build_doc_idx(indexed_dataset, documents, num_epochs, np_rng, separate_last_epoch, shuffle=True):
    sub_documents = []
    cur_ds_idx = -1
    for doc_id in documents:
        ds_idx = indexed_dataset._get_ds_idx_based_on_doc_idx(doc_id)
        if ds_idx != cur_ds_idx:
            sub_documents.append([])
            cur_ds_idx = ds_idx
        sub_documents[-1].append(doc_id)
    doc_idx_list = []
    for docs in sub_documents:
        for epoch in range(num_epochs):
            doc_idx_list.append(_build_doc_idx(docs, 1, np_rng, False, shuffle=shuffle))
    return np.concatenate(doc_idx_list)


def _costomized_build_shuffle_idx(num_samples, total_size, np_rng):
    # we will not shuffle the sample idx here
    # so just return np arange
    dtype_ = np.uint32
    if total_size >= (np.iinfo(np.uint32).max - 1):
        dtype_ = np.int64
    return np.arange(start=0, stop=total_size, step=1, dtype=dtype_)


class GPTDataset(NeMoGPTDataset):
    def __init__(
        self, 
        cfg,
        trainer, 
        tokenizer, 
        name, 
        data_prefix, 
        documents, 
        indexed_dataset, 
        num_samples, 
        seq_length, 
        seed, 
        drop_last=True,
    ):
        if isinstance(indexed_dataset, SequentialIndexedCachedDataset):
            # hack the _build_doc_idx and _build_shuffle_idx function here 
            # to make it right for SequentialIndexedCachedDataset
            import nemo.collections.nlp.data.language_modeling.megatron.gpt_dataset as nemo_gpt_dataset
            official_build_doc_idx = nemo_gpt_dataset._build_doc_idx
            official_build_shuffle_idx = nemo_gpt_dataset._build_shuffle_idx
            nemo_gpt_dataset._build_doc_idx = partial(_customized_build_doc_idx, indexed_dataset)
            nemo_gpt_dataset._build_shuffle_idx = _costomized_build_shuffle_idx
        super().__init__(
            cfg,
            trainer, 
            tokenizer, 
            name, 
            data_prefix, 
            documents, 
            indexed_dataset, 
            num_samples, 
            seq_length, 
            seed, 
            drop_last
        )
        if isinstance(indexed_dataset, SequentialIndexedCachedDataset):
            # set the hacked functions back
            nemo_gpt_dataset._build_doc_idx = official_build_doc_idx
            nemo_gpt_dataset._build_shuffle_idx = official_build_shuffle_idx


class SequentialIndexedCachedDataset(IndexedCachedDataset):
    def __init__(self, paths, cache_all=False):
        self.paths = paths
        self.inner_datasets = [IndexedCachedDataset(path) for path in paths]
        self.ds_num = len(paths)
        self.sizes = []
        sizes = [ds.sizes for ds in self.inner_datasets]
        self.sizes = np.concatenate(sizes)
        lens = [len(ds) for ds in self.inner_datasets]
        self._len = sum(lens)
        self._len_accumulate = []
        cur_len = 0
        for i in range(self.ds_num):
            self._len_accumulate.append(cur_len)
            cur_len += len(self.inner_datasets[i])
        self._len_accumulate.append(cur_len)
        self.cur_ds_idx = 0
        self.cache_all = cache_all

    def __del__(self):
        for ds in self.inner_datasets:
            del ds

    # TODO: need to implement logic to do prefetch in sub thread
    def prefetch(self, ds_idx, indices):
        if ds_idx >= self.ds_num:
            logging.warn(f"ds_idx {ds_idx} exceeds the number of inner datasets {self.ds_num}, will loop back")
            ds_idx %= self.ds_num
        indices_for_inner_ds = [i - self._len_accumulate[ds_idx] for i in indices]
        self.inner_datasets[ds_idx].prefetch(indices_for_inner_ds)

    def evict_sub_ds_cache(self, ds_idx):
        ds_idx = ds_idx % self.ds_num
        self.inner_datasets[ds_idx].cache = None
        self.inner_datasets[ds_idx].cache_idx = {}

    def _get_ds_idx_based_on_doc_idx(self, doc_idx):
        for i in range(self.ds_num):
            ds_idx = (i + self.cur_ds_idx) % self.ds_num
            if self._len_accumulate[ds] <= doc_idx < self._len_accumulate[ds_idx + 1]:
                self.cur_ds_idx = ds_idx
                return ds_idx
        raise IndexError(f'index {doc_idx} out of range {self._len_accumulate[-1]}')

    def __getitem__(self, idx):
        if isinstance(idx, int):
            ds_idx = self._get_ds_idx_based_on_doc_idx(idx)
            if not self.cache_all:
                # evict the previous ds cache
                self.evict_sub_ds_cache(ds_idx - 1)
            ds_doc_idx = idx - self._len_accumulate[ds_idx]
            return self.inner_datasets[ds_idx][ds_doc_idx]
        elif isinstance(idx, slice):
            sents = []
            for i in range(*idx.indices(len(self))):
                sents.append(self[i])
            return sents



import os
from typing import Callable, Any, Optional
from functools import wraps

from anc.api.connection import Connection
from anc.conf.remote import remote_server

training_dataset_base_default_path = "/mnt/weka/dvc_data"

def is_rank_zero() -> bool:
    node_rank = int(os.getenv("RANK", "0"))
    if node_rank == 0:
        local_rank = int(os.getenv("LOCAL_RANK", "0"))
        return local_rank == 0
    return False

def rank_zero_only(fn: Callable[..., Any]) -> Callable[..., Optional[Any]]:
    @wraps(fn)
    def wrapped_fn(*args: Any, **kwargs: Any) -> Optional[Any]:
        if is_rank_zero():
            return fn(*args, **kwargs)
        return None
    return wrapped_fn

class Dataload:
    def __init__(self, cfg):
        self._cfg = cfg
        self.dataset_name = None
        self.dataset_version = None
        self.dataset_files = None
        self.dataset_concat_sampling_probabilities = None
        if self._cfg:
            self.dataset_name = self._cfg.data.train_ds.customized_dataset.name
            self.dataset_files = self._cfg.data.train_ds.customized_dataset.files
            self.dataset_version = self._cfg.data.train_ds.customized_dataset.version
            self.dataset_concat_sampling_probabilities = self._cfg.data.train_ds.customized_dataset.concat_sampling_probabilities
    
    def _replace_cfg_data_train_ds_file_names(self, dest_path):
        if not self.dataset_files:
            return
        # reset origin dataset in cfg
        if self._cfg.data.train_ds.file_names is None:
            self._cfg.data.train_ds.file_names = []

        for file in self.dataset_files:
            self._cfg.data.train_ds.file_names.append(dest_path+"/"+file)
            
        if self.dataset_concat_sampling_probabilities:
            if self._cfg.data.train_ds.concat_sampling_probabilities is None:
                self._cfg.data.train_ds.concat_sampling_probabilities = []
            self._cfg.data.train_ds.concat_sampling_probabilities.extend(self.dataset_concat_sampling_probabilities)    

        return self._cfg

    @rank_zero_only
    def download_dataset(self, dataset_name, dataset_version, dest_path):
        # download dataset
        conn = Connection(url=remote_server)
        data = {
            "dataset_name": dataset_name,
            "version": dataset_version,
            "dest_path": dest_path
        }
        try:
            response = conn.post("/get", json=data, stream=True)
            for chunk in response.iter_lines(decode_unicode=True):
                if chunk:
                    print(chunk)
        except Exception as e:
            raise(f"Dataloader error occurred: {e}")
        print("- dataset downloaded.")

    def setup_dataset(self):
        if self.dataset_name is None or self.dataset_version is None:
            print("warning: no customized dataset/version found. Skip data setup ...")
            return
        training_dataset_base_path = os.getenv("training_dataset_base_path", training_dataset_base_default_path)
        dest_path = training_dataset_base_path + '/' + self.dataset_name + "_" + self.dataset_version
        os.makedirs(dest_path, exist_ok=True)
        # We will create a path under training_dataset_base_default_path, e.g. /mnt/weka/dvc_data/train_v1.0
        # when we finished download, we expected that the /mnt/weka/dvc_data/train_v1.0/train will be created.
        # download dataset
        self.download_dataset(self.dataset_name, self.dataset_version, dest_path)
        final_dest_path = dest_path + "/" + self.dataset_name
        updated_cfg = self._replace_cfg_data_train_ds_file_names(final_dest_path)


from pyarrow.parquet import ParquetFile
import tarfile


class DataNotReady:
    def __init__(self):
        pass


def convert_parquet_to_wds(parquet_path, wds_folder):
    import webdataset
    _, parquet_file_name = os.path.split(parquet_path)
    wds_file_name = os.path.join(wds_folder, parquet_file_name.split(".")[0])
    parquet_file = ParquetFile(parquet_path)
    num_row_groups = parquet_file.num_row_groups
    with wds.TarWriter(f"{wds_file_name}.tar") as tar:
        for rid in range(num_row_groups):
            rg = parquet_file.read_row_group(rid)
            group_data = rg.to_pandas()
            rows = group_data.to_dict('records')
            for i, row in enumerate(rows):
                print(rid, i)
                tar.write(row)
    parquet_file.close()


class TarFileReader:
    def __init__(self, path, columns=None):
        self.tar_file = tarfile.open(path)
        self.names = self.tar_file.getnames()
        if columns is None:
            columns_all = {i.rsplit(".", 1)[-1] for i in self.names}
            columns = list(columns_all)
        self.columns = columns

    def __getitem__(self, key):
        targets = [f"{key}.{column}" for column in self.columns]
        tarinfos = [self.tar_file.getmember(name) for name in targets]
        tarinfos.sort(key=lambda x: x.offset)
        data = {}
        for tarinfo in tarinfos:
            cur_data = self.tar_file.extractfile(tarinfo).read()
            data[tarinfo.name[len(key)+1:]] = cur_data
        return data
    def close(self):
        self.tar_file.close()


def test_parquet(parquet_path):
    t1 = time.time()
    parquet_file = ParquetFile(parquet_path)
    num_row_groups = parquet_file.num_row_groups
    for rid in range(num_row_groups):
        rg = parquet_file.read_row_group(rid)
        group_data = rg.to_pandas()
        rows = group_data.to_dict('records')
    parquet_file.close()
    print(f"parquet test time: {time.time() - t1}s")


def test_tar(tar_file, shuffle=False):
    t1 = time.time()
    r = TarFileReader(tar_file)
    keys = [i.split(".")[0] for i in r.names if i.endswith(".wav")]
    if shuffle:
        random.shuffle(keys)
    for k in keys:
        d = r[k]
    r.close()
    print(f"tar test time: {time.time() - t1}s")



import click

from .util import (
    console,
    click_group,
)


@click_group()
def deployment():
    pass


def add_command(cli_group):
    cli_group.add_command(deployment)



@deployment.command(name="list")
@click.option(
    "--pattern",
    "-p",
    help="Regular expression pattern to filter deployment names.",
    default=None,
)
def list_command(pattern):
    print("hello my pattern", pattern)



import os
import time
from torch.utils.data import DataLoader
from anc.data.anc_dataset import AncDataset
from anc.data.anc_sampler import AncSampler
from anc.data.anc_processor import Processor, AncProcessor
from anc.data.utils import DataNotReady
import logging
import datetime


class AncDataLoader:
    r'''
    A wrapper for dataloader, dataset and sampler
    It makes the handling of the data gathering logic in each subprocess easier
    It also helps handle the checkpoint load and resume logic

    Args:
        paths (list of str): files from which to load the data.
        batch_size (int): how many samples per batch to load.
        num_workers (int): how many subprocesses to use for data
            loading. ``0`` means that the data will be loaded in the main process.
        rank (int, optional): data parallel rank of current process (default: ``0``).
        world (int, optional): data parallel world size (default: ``1``).
        processor (Processor, optional): object that defines the data transform and batch transform
            logic (default ``AncProcessor`` instance which just return the input as is for both
            transform and batch transform).
        data_type (str, optional): the type of dataset that the paths represents (default ``parquet``).
        shuffle (bool, optional): set to ``True`` to have the data reshuffled
            at every epoch (default: ``False``).
        drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
            if the dataset size is not divisible by the batch size. If ``False`` and
            the size of dataset is not divisible by the batch size, then the last batch
            will be smaller. (default: ``False``).
        seed (int, optional): seed for randomness (default: ``0``).
        pin_memory (bool, optional): If ``True``, the data loader will copy Tensors
            into device/CUDA pinned memory before returning them.  If your data elements
            are a custom type, or your :attr:`collate_fn` returns a batch that is a custom type,
            see the example below.
        ds_args (dict, optional): dataset arguments for data loading, such as the columns to load 
            for parquet files, etc (default ``{}``).
        is_train (bool, optional): whether this dataset if for training (default ``True``).
        data_borrow (bool, optional): If ``True``, worker would borrow data from peers when its data is not ready.
            This would do good to performance, but make the bitwise checkpoint harder (default ``True``).
        repeat (bool, optional): If ``True``, data would be loaded repeatedly
        prefetch_factor (int, optional): the prefetch_factor for torch dataloader
    '''
    def __init__(
        self,
        paths,
        batch_size,
        num_workers,
        rank=0,
        world=1,
        processor=AncProcessor(),
        data_type="parquet",
        shuffle=False,
        drop_last=False,
        seed=0,
        pin_memory=True,
        ds_args={},
        is_train=True,
        data_borrow=False,
        repeat=False,
        prefetch_factor=2,
    ):
    # TODO: add support to get rank and world from torch env
        self.paths = paths
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.dataset = AncDataset(paths, processor=processor, dataset_type=data_type, ds_args=ds_args)
        self.sampler = AncSampler(self.dataset, batch_size, world, rank, num_workers, shuffle, seed, drop_last, repeat=repeat)
        self.dataset.set_sampler(self.sampler)
        if data_borrow:
            assert repeat, "repeat must be true if data_borrow is set"
        try:
            self.loader = DataLoader(
                self.dataset,
                num_workers=num_workers,
                pin_memory=pin_memory,
                collate_fn=lambda x: x,
                prefetch_factor=prefetch_factor,
                enable_borrow=data_borrow
            )
        except:
            self.loader = DataLoader(
                self.dataset,
                num_workers=num_workers,
                pin_memory=pin_memory,
                collate_fn=lambda x: x,
                prefetch_factor=prefetch_factor
            )
            logging.warning("Dataloader does not have enable_borrow attribute!")
        self.step = 0
        self.rank = rank
        self.world = world
        self.loader_iter = None
        self.buffer = [[] for _ in range(num_workers)]
        self.last_wid = -1
        self.out_of_data_worker = set()
        self.is_train = is_train
        self.data_borrow = data_borrow

    def __iter__(self):
        logging.debug(f"pid {os.getpid()} rank {self.rank} in iter")
        self.loader_iter = iter(self.loader)
        self.out_of_data_worker = set()
        return self
    
    def __len__(self):
        return len(self.sampler)
    
    def _get_borrow_id(self):
        max_buffer_id = -1
        max_buffer_length = 0
        for i in range(self.num_workers):
            if len(self.buffer[i]) > max_buffer_length:
                max_buffer_length = len(self.buffer[i])
                max_buffer_id = i
        return max_buffer_id

    def __next__(self):
        now = datetime.datetime.now()
        need_logging = self.loader_iter._data_queue.qsize() < 10
        t1 = time.time()
        if len(self.out_of_data_worker) == self.num_workers:
            # TODO: better handle the exception
            raise StopIteration
        # decide which subprocess to get data from
        if self.last_wid < 0:
            cur_wid = self.step % self.num_workers
        else:
            cur_wid = (self.last_wid + 1) % self.num_workers
        if cur_wid in self.out_of_data_worker:
            for i in range(self.num_workers):
                cur_wid = (cur_wid + 1) % self.num_workers
                if cur_wid not in self.out_of_data_worker:
                    break
        data = None
        if len(self.buffer[cur_wid]) > 0:
            data = self.buffer[cur_wid].pop()
        else:
            cnt = 0
            while True:
                item = next(self.loader_iter)
                # since we set batch size=1 for torch dataloader, 
                # the actual batch size is handled by sampler
                # we need to use index 0 to get data out
                # note: cur_data here should be a list of batch data
                cur_data_wid, cur_data, is_last_batch = item[0]
                if is_last_batch:
                    self.out_of_data_worker.add(cur_data_wid)
                if isinstance(cur_data, DataNotReady):
                    logging.info(f"next call get DataNotReady from {cur_data_wid}, is last batch {is_last_batch}")
                    continue
                assert isinstance(cur_data, list)
                if cur_data_wid == cur_wid:
                    data = cur_data[0]
                    self.buffer[cur_data_wid] += cur_data[1:]
                    break
                else:
                    # store the out of order data
                    self.buffer[cur_data_wid] += cur_data
                cnt += 1
                if cnt % self.num_workers == 0:
                    logging.warning(f"{self.rank} Too many loop ({cnt} times) when getting next data for worker {cur_wid} of step {self.step}")
        self.step += 1
        self.last_wid = cur_wid
        consumed_time = time.time() - t1
        need_logging = need_logging or consumed_time > 0.01
        if need_logging:
            logging.info(f"__next__ called time {now} pid {os.getpid()} rank {self.rank} return step {self.step} consumed time {consumed_time} data with buffer {self.loader_iter._data_queue.qsize()}")
        return data

    def __getstate__(self):
        state = {}
        state['paths'] = self.paths
        state['batch_size'] = self.batch_size
        state['num_workers'] = self.num_workers
        state['sampler'] = self.sampler.__getstate__()
        # TODO: add actual subprocess's state
        state['dataset'] = self.dataset.__getstate__()
        state['last_wid'] = self.last_wid
        state['out_of_data_worker'] = self.out_of_data_worker
        state['step'] = self.step

    def __setstate__(self, state):
        if 'dataset' in state:
            dataset_state = state.pop('dataset')
            self.dataset.__setstate__(dataset_state)
        if 'sampler' in state:
            sampler_state = state.pop('sampler')
            self.sampler.__setstate__(sampler_state)
        self.__dict__.update(state)


if __name__ == "__main__":
    folder = "/mnt/personal/parquet_demo_data"
    fnames = [
        "01_0001.parquet",
        "02_0001.parquet",
        "03_0001.parquet",
        "04_0001.parquet",
        "05_0001.parquet",
        "06_0001.parquet",
        "07_0001.parquet"
    ]
    files = [f"{folder}/{fname}" for fname in fnames]
    loader = AncDataLoader(files, 1024, num_workers=4)
    for data in loader:
        print(len(data))



import sys
import os
from typing import Any, Optional
from urllib.parse import urlparse

import click
import yaml
from typing import Dict

from rich.console import Console

from anc.conf.remote import remote_storage_prefix

console = Console(highlight=False)


def click_group(*args, **kwargs):
    class ClickAliasedGroup(click.Group):
        def get_command(self, ctx, cmd_name):
            rv = click.Group.get_command(self, ctx, cmd_name)
            if rv is not None:
                return rv

            def is_abbrev(x, y):
                # first char must match
                if x[0] != y[0]:
                    return False
                it = iter(y)
                return all(any(c == ch for c in it) for ch in x)

            matches = [x for x in self.list_commands(ctx) if is_abbrev(cmd_name, x)]

            if not matches:
                return None
            elif len(matches) == 1:
                return click.Group.get_command(self, ctx, matches[0])
            ctx.fail(f"'{cmd_name}' is ambiguous: {', '.join(sorted(matches))}")

        def resolve_command(self, ctx, args):
            # always return the full command name
            _, cmd, args = super().resolve_command(ctx, args)
            return cmd.name, cmd, args

    return click.group(*args, cls=ClickAliasedGroup, **kwargs)

def is_valid_source_path(path: str, personal: str) -> bool:
    """Check if the source path is valid."""
    if not os.path.exists(path):
        print(f"Your source {path} is invalid, path not exists")
        return False
    if not os.access(path, os.R_OK):
        print(f"Your source {path} is invalid, path not access")
        return False
    if not path.startswith(remote_storage_prefix):
        print(f"Your source {path} is invalid, path is not prefix of {remote_storage_prefix} ")
        return False
    if path.startswith("/mnt/personal") and not personal:
        print(f"We can't get your personal information as you want to use {path}, so please reach out infra team to setup.")
        return False
    return True

def convert_to_absolute_path(path: str) -> str:
    """Convert a relative path to an absolute path."""
    return os.path.abspath(path)

def get_file_or_folder_name(path: str) -> str:
    """Get the file name with extension or folder name from the given path."""
    if os.path.isdir(path):
        return os.path.basename(path)  # Return folder name
    elif os.path.isfile(path):
        return os.path.basename(path)  # Return file name with extension
    else:
        raise ValueError(f"Invalid path {path}")


class ConfigManager:
    """Configuration management class that handles priority between environment variables and local config files"""

    def __init__(self):
        # Define environment variable names
        self.ENV_USER = "MLP_USER"
        self.ENV_PROJECT = "MLP_PROJECT"
        self.ENV_CLUSTER = "MLP_CLUSTER"
        self.cluster_config_path = '/mnt/project/.anc_profile'
        self.personal_config_path = '~/.anc_personal'

    def _read_anc_cluster_profile(self):
        """Read cluster configuration file
        Returns:
            dict: Cluster configuration dictionary
        """
        # Check if .anc_profile file exists
        if not os.path.isfile(self.cluster_config_path):
            return {}

        # Read and parse the YAML file
        try:
            with open(self.cluster_config_path, 'r') as file:
                profile_data = yaml.safe_load(file)
            return profile_data if profile_data else {}
        except yaml.YAMLError as e:
            return {}
        except IOError as e:
            return {}

    def _read_anc_personal_profile(self):
        """Read personal configuration file
        Returns:
            dict: Personal configuration dictionary
        """
        # Implement actual file reading logic here
        # e.g. personal: xuguang.zhao
        # Check if .anc_profile file exists
        profile_path = os.path.expanduser(self.personal_config_path)

        if not os.path.isfile(profile_path):
            return {}

        try:
            with open(profile_path, 'r') as file:
                profile_data = yaml.safe_load(file)
            return profile_data or {}
        except yaml.YAMLError as e:
            return {}
        except IOError as e:
            return {}

    def get_environment(self):
        """Get environment configuration information
        Priority: First try to get from environment variables,
        if not found, then read from configuration files

        Returns:
            tuple: (project, cluster, personal) configuration values
        """
        # Get values from environment variables
        cluster_info = self._read_anc_cluster_profile()
        user_info = self._read_anc_personal_profile()

        personal = os.environ.get(self.ENV_USER) or user_info.get('personal', '')
        cluster = os.environ.get(self.ENV_CLUSTER) or cluster_info.get('cluster', '')
        project = os.environ.get(self.ENV_PROJECT) or cluster_info.get('project', '')
            
        return project, cluster, personal



from pyarrow.parquet import ParquetFile
import multiprocessing.dummy as mt
import os


def get_inner_group_idx_from_row_idx(rows_accumulate, row_idx, curr_inner_group_idx, inner_group_num):
    assert curr_inner_group_idx < inner_group_num
    if rows_accumulate[curr_inner_group_idx] <= row_idx < rows_accumulate[curr_inner_group_idx + 1]:
        return curr_inner_group_idx
    for i in range(inner_group_num):
        cur_gidx = (curr_inner_group_idx + i + 1) % inner_group_num
        if rows_accumulate[cur_gidx] <= row_idx < rows_accumulate[cur_gidx + 1]:
            return cur_gidx
    return -1


class ParquetDataset:
    r'''
    Given a path of parquet file, return target data by index
    Note: This class only implements an efficient way to return sample by index
        The index order is controlled by sampler, not here.

    Args:
        filepath (str or tuple of str): file from which to load the data. If is tuple of str, each 
            each str represents a single parquet file, and corresponding rows across those files together
            form a sample
        columns (list of str, optional), columns to load (default: ``None`` means load all columns)
    '''
    def __init__(self, filepath, columns=None):

        self.filepath = filepath
        if isinstance(filepath, tuple) or isinstance(filepath, list):
            self.parquet_file = [ParquetFile(path) for path in filepath if os.path.exists(path)]
        else:
            self.parquet_file = [ParquetFile(filepath),]
        self.num_row_groups = self.parquet_file[0].num_row_groups
        self.row_groups = list(range(self.num_row_groups))
        self.row_group_rows = [self.parquet_file[0].metadata.row_group(i).num_rows for i in self.row_groups]

        self.row_group_rows_accumulate = []
        rows = 0
        for i in self.row_groups:
            self.row_group_rows_accumulate.append(rows)
            rows += self.row_group_rows[i]
        self.row_group_rows_accumulate.append(rows)
        self.cur_row_group = 0

        self.num_rows = self.parquet_file[0].metadata.num_rows
        self.columns = columns
        self.cache = {}
        # ensure that the metadata of the group files is same
        for i in range(1, len(self.parquet_file)):
            assert self.num_row_groups == self.parquet_file[i].num_row_groups
            assert self.row_group_rows == [self.parquet_file[i].metadata.row_group(rid).num_rows for rid in self.row_groups]
    
    def _read_row_group(self, row_group):
        rows = []
        for pf in self.parquet_file:
            row_group_data = pf.read_row_group(row_group, columns=self.columns)
            group_data = row_group_data.to_pandas()
            cur_rows = group_data.to_dict('records')
            if len(rows) == 0:
                rows = cur_rows
            else:
                assert len(rows) == len(cur_rows), "row group of each file should be the same"
                for i in range(len(rows)):
                    rows[i].update(cur_rows[i])
        return rows

    def __iter__(self):
        for row_group in self.row_groups:
            rows = self._read_row_group(row_group)
            yield from rows
    
    def _get_row_group_idx_from_ridx(self, ridx):
        if self.row_group_rows_accumulate[self.cur_row_group] <= ridx < self.row_group_rows_accumulate[self.cur_row_group + 1]:
            return self.cur_row_group
        for i in self.row_groups:
            cur_gidx = (self.cur_row_group + i) % self.num_row_groups
            if self.row_group_rows_accumulate[cur_gidx] <= ridx < self.row_group_rows_accumulate[cur_gidx + 1]:
                self.cur_row_group = cur_gidx
                return self.cur_row_group
        return -1
    
    def _clean_cache(self):
        # TODO: maybe a better way to find out which row groups to delete
        gidx_to_delete = [i for i in self.cache]
        for i in gidx_to_delete:
            del self.cache[i]
    
    def __getitem__(self, idx):
        assert 0 <= idx < self.num_rows, f"idx shoud be in [0, {self.num_rows}) but get {idx}"
        cur_row_group = get_inner_group_idx_from_row_idx(
            self.row_group_rows_accumulate, idx, self.cur_row_group, self.num_row_groups)
        assert cur_row_group >= 0, f"invalid row group {cur_row_group} for idx {idx}"
        self.cur_row_group = cur_row_group
        if cur_row_group not in self.cache:
            cur_rows = self._read_row_group(cur_row_group)
            self._clean_cache()
            self.cache[cur_row_group] = cur_rows
        else:
            cur_rows = self.cache[cur_row_group]
        return cur_rows[idx - self.row_group_rows_accumulate[cur_row_group]]
    
    def __len__(self):
        return self.num_rows

    def close(self):
        del self.cache
        for pf in self.parquet_file:
            pf.close()
    
    def get_sub_lengths(self):
        return self.row_group_rows


class ParquetConcateDataset:
    r'''
    Given multiple parquet files as a whole dataset, return target data by index
    Note: This class only implements an efficient way to return sample by index.
        The index order is controlled by sampler, not here.

    Args:
        filepaths (list of str): files from which to load the data.
        columns (list of str, optional), columns to load (default: ``None`` means load all columns)
    '''
    def __init__(self, filepaths, columns=None):
        self.filepaths = filepaths
        self.num_files = len(filepaths)
        self.columns = columns
        rows = []
        rows_per_group = []
        with mt.Pool(16) as p:
            sub_ds_info = p.map(self._get_sub_ds_length, self.filepaths)
        rows = [i[0] for i in sub_ds_info]
        rows_per_group = [i[1] for i in sub_ds_info]
        self.rows_per_file = rows
        self.rows_per_group = rows_per_group
        self.num_rows = sum(self.rows_per_file)
        self.num_rows_accumulate = []
        cur_rows = 0
        for i in range(self.num_files):
            self.num_rows_accumulate.append(cur_rows)
            cur_rows += rows[i]
        self.num_rows_accumulate.append(cur_rows)
        self.cur_ds_idx = 0
        self.cache = {}
    
    def _get_sub_ds_length(self, path):
        ds = ParquetDataset(path)
        row = len(ds)
        row_groups = ds.get_sub_lengths()
        ds.close()
        return row, row_groups

    def __len__(self):
        return self.num_rows
    
    def _clean_cache(self):
        ds_idx_to_delete = [i for i in self.cache]
        for i in ds_idx_to_delete:
            self.cache[i].close()
            del self.cache[i]

    def __getitem__(self, idx):
        assert 0 <= idx < self.num_rows, f"idx shoud be in [0, {self.num_rows}) but get {idx}"
        cur_ds_idx = get_inner_group_idx_from_row_idx(
            self.num_rows_accumulate, idx, self.cur_ds_idx, self.num_files)
        assert cur_ds_idx >= 0, f"invalid file idx {cur_row_group} for row idx {idx}"
        self.cur_ds_idx = cur_ds_idx
        if cur_ds_idx not in self.cache:
            cur_ds = ParquetDataset(self.filepaths[cur_ds_idx], self.columns)
            self._clean_cache()
            self.cache[cur_ds_idx] = cur_ds
        else:
            cur_ds = self.cache[cur_ds_idx]
        data = cur_ds[idx - self.num_rows_accumulate[cur_ds_idx]]
        if 'filepath' not in data:
            data['filepath'] = self.filepaths[cur_ds_idx]
        return data
    
    def get_sub_lengths(self, level="row_group"):
        assert level in {"row_group", "file"}
        if level == "row_group":
            return self.rows_per_group
        else:
            return self.rows_per_file


if __name__ == "__main__":
    folder = "/mnt/personal/parquet_demo_data"
    fnames = [
        "01_0001.parquet",
        "02_0001.parquet",
        "03_0001.parquet",
        "04_0001.parquet",
        "05_0001.parquet",
        "06_0001.parquet",
        "07_0001.parquet"
    ]
    files = [f"{folder}/{fname}" for fname in fnames]
    ds = ParquetConcateDataset(files)
    assert [sum(i) for i in ds.rows_per_group] == ds.rows_per_file
    assert sum(ds.rows_per_file) == len(ds)
    for i in range(10):
        data = ds[i * 500]
        print(data['filepath'])



import click
from anc.version import __version__

from anc.cli import deployment
from anc.cli import dataset
from anc.cli.util import click_group

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.version_option(__version__, "-v", "--version")
@click_group(context_settings=CONTEXT_SETTINGS)
def main():
    pass


#eployment.add_command(main)
dataset.add_command(main)

if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod


class Processor(ABC):
    r'''
    function to handle a single data
    return None if no data needs to return. Otherwise, return a list of transformed data
    It is useful when certain filtering logic is needed
    '''
    @abstractmethod
    def transform(self, item, is_last_sample=False):
        pass

    r'''
    function to handle a batch of data
    input:
        list_of_items -- data to be batch processed
        is_last_batch -- whether current batch is the last batch might be useful if the last batch's 
                         tranform logic is different from the normal batches.
    return:
        the transformed batch data. None if no data needs to return, otherwise, it should be a list of batch data
        since buffer gathering logic might exist, and it is possible to return multiple of batch data.
    '''
    @abstractmethod
    def batch_transform(self, list_of_items, is_last_batch=False):
        pass


class AncProcessor(Processor):
    r'''
    A processor that does nothing but return the input as is for both transform and batch transform function
    '''
    def transform(self, item, is_last_sample=False):
        return [item]

    # we wrap out data to a list to follow the return pattern in the base class
    def batch_transform(self, list_of_items, is_last_batch=False):
        return [list_of_items]


import requests
from typing import Optional, Dict

class Connection:
    def __init__(self, url: str, token: Optional[str] = None):
        self._url = url
        self._token = token
        self._header = {"Authorization": "Bearer " + token} if token else {}
        self._timeout = 120
        self._session = requests.Session()

    def get(self, path: str, *args, **kwargs):
        return self._session.get(self._url + path, *args, **kwargs)

    def post(self, path: str, *args, stream: bool = False, **kwargs):
        response = self._session.post(self._url + path, *args, **kwargs, stream=stream)
        return response

    def patch(self, path: str, *args, **kwargs):
        return self._session.patch(self._url + path, *args, **self._safe_add(kwargs))

    def put(self, path: str, *args, **kwargs):
        return self._session.put(self._url + path, *args, **self._safe_add(kwargs))

    def delete(self, path: str, *args, **kwargs):
        return self._session.delete(self._url + path, *args, **self._safe_add(kwargs))

    def head(self, path: str, *args, **kwargs):
        return self._session.head(self._url + path, *args, **self._safe_add(kwargs))



from .util import is_valid_source_path, get_file_or_folder_name, convert_to_absolute_path
from anc.conf.remote import remote_server, remote_storage_prefix, repo_url
from anc.api.connection import Connection
import subprocess
import os
import sys
import json
from requests.exceptions import RequestException
from tabulate import tabulate

from .util import ConfigManager


class DatasetOperator:
    def __init__(self):
        self.conn = Connection(url=remote_server)
        self.config_manager = ConfigManager()

    def download_dataset(self, name, version, dest, cache_policy):
        git_commit_id = self.get_commit_id(name, version)
        if git_commit_id is None:
            print("Not found a git commit id for the dataset, can't do download.")
            return
        dest = os.path.abspath(dest)
        if remote_storage_prefix in dest:
            return self.remote_download(name, version, dest, cache_policy)
        self.local_download(git_commit_id, name, dest)

    # for training job use.
    def add_dataset(self, dataset_name, version, source_path, message):
        if remote_storage_prefix in source_path:
            return self.remote_add(dataset_name, version, source_path, message)
        self.local_add(dataset_name, version, source_path, message)

    def remote_add(self, version, source_path, message):
        project, cluster, peronsal = self.config_manager.get_environment()
        if not project or not cluster:
            print("Failed to add dataset, because in current enviroment cli can't identify your project or cluster, plz reach out infra team to setup")
            sys.exit(1)
        source_path = os.path.abspath(source_path)
        if not is_valid_source_path(source_path, peronsal):
            sys.exit(1)
        abs_path = convert_to_absolute_path(source_path)
        dataset_name = get_file_or_folder_name(abs_path)
        conn = Connection(url=remote_server)
        data = {
            "dataset_name": dataset_name,
            "version": version,
            "source_path": abs_path,
            "dest_path": "local",
            "project": project,
            'cluster': cluster,
            'personal': peronsal,
            "message": message
        }
        try:
            response = conn.post("/add", json=data)

            # Check if the status code is in the 2xx range
            if 200 <= response.status_code < 300:
                response_data = response.json()
                task_id = response_data.get('task_id')
                if task_id:
                    print(f"Task added successfully. Your task ID is: \033[92m{task_id}\033[0m")
                    print(f"You can check the status later by running:  \033[92manc ds list -n {dataset_name}\033[0m")
                    print(f"You can check the task status:  \033[92manc ds task status {task_id}\033[0m")
                    print(f"If your task has been pending a long time, please check the queue with: anc ds queue status")
                else:
                    print("Task added successfully, but no task ID was returned.")
            else:
                #print(f"Error: Server responded with status code {response.status_code}")
                print(f"{response.text}")
  
        except RequestException as e:
            print(f"Sorry, you can't add dataset out of clusters, please use it in a notebook")
        except json.JSONDecodeError:
            print("Sorry: received invalid JSON response from server")
        except KeyboardInterrupt:
            print(f"Operation interrupted.")
            sys.exit(0)
        except Exception as e:
            print(f"Sorry, your command run failed, you can try again or reach out infra team")

    # TODO
    def local_add(self, dataset_name, version, source_path, message):
        if os.getenv("AWS_ACCESS_KEY_ID", None) is None or os.getenv("AWS_SECRET_ACCESS_KEY", None) is None or os.getenv("GIT_TOKEN", None) is None:
            print("!!!! Hey, are you sure you want to upload it to remote? if yes, please set environment for <AWS_ACCESS_KEY_ID> and <AWS_SECRET_ACCESS_KEY> and <GIT_TOKEN>, so we can continue !!!!")
            sys.exit(1)


    def remote_download(self, name, version, dest, cache_policy):
        abs_path = convert_to_absolute_path(dest)
        project, cluster, personal = self.config_manager.get_environment()
        if not project or not cluster:
            print("Failed to download dataset, because in current enviroment cli can't identify your project or cluster, plz reach out infra team to setup")
            sys.exit(1)

        if abs_path.startswith("/mnt/personal") and not personal:
            print(f"We can't get your personal information as you want to use {abs_path}, so please reach out infra team to setup.")
            return False
        data = {
            "dataset_name": name,
            "version": version,
            "dest_path": abs_path,
            "cache_policy": cache_policy,
            "project": project,
            "cluster": cluster,
            "personal": personal,
        }
        try:
            response = self.conn.post("/get", json=data)

            # Check if the status code is in the 2xx range
            if 200 <= response.status_code < 300:
                response_data = response.json()
                task_id = response_data.get('task_id')
                if task_id:
                    print(f"Dataset get operation initiated. Your task ID is: \033[92m{task_id}\033[0m")
                    print(f"You can check the status later by running: \033[92manc ds task status {task_id}\033[0m")
                    print(f"If your task has been pending a long time, please check the queue with: \033[92manc ds queue status\033[0m")
                    print(f"please don't do any directory change along with {abs_path}")
                else:
                    print("Dataset get operation initiated, but no task ID was returned.")
            else:
                print(f"Error: Server responded with status code {response.status_code}")
                print(f"{response.text}")

        except RequestException as e:
            print(f"Sorry, you can't add dataset out of clusters, please use it in a notebook")
        except json.JSONDecodeError:
            print("Error: Received invalid JSON response from server")
        except KeyboardInterrupt:
            print(f"Operation interrupted.")
            sys.exit(0)
        except Exception as e:
            print(f"Sorry, your command run failed, you can try again or reach out infra team")

    def list_dataset(self, dataset_name):
        response = self.conn.get("/query_datasets", params={"dataset_name": dataset_name})
        if response.status_code == 200:
            data = response.json()
            headers = [
                "Created At", "Dataset Name", 
                "Dataset Version",  "Message"
            ]
            table = [
                [
                    item["created_at"], item["dataset_name"],
                    item["dataset_version"],
                    item["message"]
                ] for item in data
            ]
            print(tabulate(table, headers=headers, tablefmt="grid", disable_numparse=True))
        else:
            print("Failed to retrieve datasets. Status code:", response.status_code)


    def get_commit_id(self, dataset_name, dataset_version):
        data = {
            "dataset_name": dataset_name,
            "version": dataset_version,
        }
        try:
            response = self.conn.get("/query_datasets", params={"dataset_name": dataset_name, "dataset_version": dataset_version})
            data = response.json()
            return data[0]["git_commit_id"] if len(data) > 0 and "git_commit_id" in data[0] else None
        except Exception as e:
            print(f"Error occurred: {e}")


    def local_download(self, git_commit_id, dataset_name, dest_path):
        print("<<This is a local download operation>>")
        if os.getenv("AWS_ACCESS_KEY_ID", None) is None or os.getenv("AWS_SECRET_ACCESS_KEY", None) is None or os.getenv("GIT_TOKEN", None) is None:
            print("!!!! Hey, are you sure you want to download it to your local? if yes, please set environment for <AWS_ACCESS_KEY_ID> and <AWS_SECRET_ACCESS_KEY> and <GIT_TOKEN>, so we can continue !!!!")
            sys.exit(1)
        url = repo_url
        git_token = os.environ["GIT_TOKEN"]
        repo_url_with_token = url.replace("https://", f"https://{git_token}@")
        source_path = 'local/' +  dataset_name
        command = ["dvc", "get", repo_url_with_token, source_path, "--rev", git_commit_id, "-o", dest_path]

        try:
            subprocess.run(command, check=True)
            print(f"Successfully downloaded {source_path} from {repo_url} to {dest_path}.")
        except subprocess.CalledProcessError as e:
            raise ValueError(f"Failed to download {source_path}: {e}")



import toml
from pathlib import Path

def get_version():
    pyproject_path = Path(__file__).parent.parent / "pyproject.toml"
    pyproject_content = toml.load(pyproject_path)
    return pyproject_content["tool"]["poetry"]["version"]

__version__ = get_version()



from torch.utils.data.sampler import *
from torch.utils.data.distributed import DistributedSampler
import random
import time


class AncSampler(DistributedSampler):
    r'''
    A batch sampler controls the data sharding, data shuffling and data loading order with proper indices

    The data sharding is chunk based. For example, to shard a dataset with 10 elements into 2 splits, 
    the result data index would be [[0,1,2,3,4],[5,6,7,8,9]] instead of [[0,2,4,6,8],[1,3,5,7,9]]

    Args:
        dataset: dataset from which to load the data.
        batch_size (int): how many samples per batch to load.
        world (int, optional): data parallel world size (default: ``1``).
        rank (int, optional): data parallel rank of current process (default: ``0``).
        num_workers (int): how many subprocesses to use for data
            loading. ``0`` means that the data will be loaded in the main process.
        shuffle (bool, optional): set to ``True`` to have the data reshuffled
            at every epoch (default: ``False``).
        drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
            if the dataset size is not divisible by the batch size. If ``False`` and
            the size of dataset is not divisible by the batch size, then the last batch
            will be smaller. (default: ``False``).
        seed (int, optional): seed for randomness (default: ``0``).
        resume_step (int, optional): the step to resume from, 
            the previous steps will be skipped (default: ``0``).
        repeat (bool, optional): set to ``True`` to repeat the indices when gone through 
            all the data. If ``False`` than StopIteration will be raised when all data 
            is consumed (default: ``False``).
    '''
    def __init__(
        self,
        dataset,
        batch_size,
        world = 1,
        rank = 0,
        num_workers = 1,
        shuffle = False,
        drop_last = False,
        seed = 0,
        resume_step = 0,
        repeat = False,
    ):
        super().__init__(dataset, world, rank, shuffle, seed, drop_last)
        self.sub_lengths = dataset.get_sub_lengths()
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.resume_step = resume_step
        self.step = resume_step
        self.seed = seed
        self.last_iter_epoch = -1
        self.indices = None
        self.repeat = repeat
        self.inner_epoch_count = 0
    
    def _get_indices(self, sub_lengths, offset=0, return_sub_indices=False):
        indices_list = []
        random.seed(self.seed)
        accumulate_length = offset
        for item in sub_lengths:
            sub_indices = list(range(accumulate_length, accumulate_length + item))
            if self.shuffle:
                random.shuffle(sub_indices)
            indices_list.append(sub_indices)
            accumulate_length += item

        if self.shuffle:
            random.shuffle(indices_list)
        if return_sub_indices:
            return indices_list
        indices = []
        for item in indices_list:
            indices += item
        return indices

    def create_chunk_indices_from_bz_and_worker(self, indices):
        # calculate num batches per worker
        num_batches = (self.num_samples + self.batch_size - 1) // self.batch_size
        last_batch_size = self.num_samples - (num_batches - 1) * self.batch_size
        num_batches_per_worker = num_batches // self.num_workers
        remain = num_batches - self.num_workers * num_batches_per_worker
        num_batches_per_each_worker = [num_batches_per_worker + 1] * remain + [num_batches_per_worker] * (self.num_workers - remain)
        assert sum(num_batches_per_each_worker) == num_batches
        last_batch_worker_id = self.num_workers - 1 if remain == 0 else remain - 1

        if self.repeat and self.shuffle:
            random.seed(self.seed)
        while True:
            self.inner_epoch_count += 1
            indices_per_worker = []
            offset = 0
            for i in range(self.num_workers):
                cur_worker_indices = []
                for batch_cnt in range(num_batches_per_each_worker[i]):
                    cur_batch_size = self.batch_size
                    if i == last_batch_worker_id and batch_cnt == num_batches_per_each_worker[i] - 1:
                        cur_batch_size = last_batch_size
                    cur_worker_indices.append(indices[offset: offset + cur_batch_size])
                    offset += cur_batch_size
                indices_per_worker.append(cur_worker_indices)

            for i in range(num_batches_per_worker):
                for wid in range(self.num_workers):
                    is_last_batch = wid >= remain and i == num_batches_per_each_worker[wid] - 1 and not self.repeat
                    yield indices_per_worker[wid][i], is_last_batch
            for wid in range(remain):
                is_last_batch = not self.repeat
                yield indices_per_worker[wid][-1], is_last_batch
            if not self.repeat:
                break
            if self.shuffle:
                shift_offset = random.randint(0, len(indices))
                indices = indices[shift_offset:] + indices[:shift_offset]

    def __iter__(self):
        indices = []
        if self.epoch == self.last_iter_epoch:
            indices = self.indices
        else:
            self.last_iter_epoch = self.epoch
            if isinstance(self.sub_lengths[0], list):
                sub_indices_list = []
                accumulate_length = 0
                for item in self.sub_lengths:
                    sub_indices = self._get_indices(item, accumulate_length, True)
                    accumulate_length += sum([len(i) for i in sub_indices])
                    sub_indices_list += sub_indices
                if self.shuffle:
                    random.shuffle(sub_indices_list)
                for item in sub_indices_list:
                    indices += item
            else:
                indices = self._get_indices(self.sub_lengths)
            
            if not self.drop_last:
                # add extra samples to make it evenly divisible
                padding_size = self.total_size - len(indices)
                if padding_size <= len(indices):
                    indices += indices[:padding_size]
                else:
                    indices += (indices * math.ceil(padding_size / len(indices)))[
                        :padding_size
                    ]
            else:
                # remove tail of data to make it evenly divisible.
                indices = indices[: self.total_size]
            assert len(indices) == self.total_size

            # we do the chunk sharding here
            indices = indices[self.rank * self.num_samples : (self.rank + 1) * self.num_samples]
            self.indices = indices
        assert len(indices) == self.num_samples
        # further split the per rank indices into num_workers splits
        indices_generator = self.create_chunk_indices_from_bz_and_worker(indices)

        # skip the indices that already been consumed
        for i in range(self.resume_step):
            _ = next(indices_generator)

        # yield empty indices to let workers follow the original order
        # otherwise, the first generated indices would be consumed by worker 0
        # while they should be consumed by cur_wid
        cur_wid = self.resume_step % self.num_workers
        for i in range(cur_wid):
            yield [], False

        while True:
            try:
                yield next(indices_generator)
                self.step += 1
            except StopIteration:
                break

    def set_step(self, step):
        self.resume_step = step
        self.step = step
    
    def __len__(self):
        # this is a batch sampler, return the number of batches
        if not self.repeat:
            return (self.num_samples + self.batch_size - 1) // self.batch_size
        else:
            return 1000000000000

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getstate__(self):
        state = {}
        state['resume_step'] = self.step
        state['step'] = self.step
        state['epoch'] = self.epoch
        state['seed'] = self.seed
        state['inner_epoch_count'] = self.inner_epoch_count


if __name__ == "__main__":
    class SimpleDataset:
        def __init__(self, sub_ds_count, sub_ds_size):
            self.sub_ds_count = sub_ds_count
            self.sub_ds_size = sub_ds_size
        def get_sub_lengths(self):
            return [self.sub_ds_size] * self.sub_ds_count
        def __len__(self):
            return self.sub_ds_count * self.sub_ds_size

    def check_sampler_indices(shuffle, batch_size, resume_step):
        # create a fake dataset with 3 files and each file contains 100 elements
        sub_ds_size = 40
        sub_ds_count = 3
        ds = SimpleDataset(sub_ds_count, sub_ds_size)
        world = 4
        rank = 0
        num_workers = 2
        drop_last = False
        num_per_rank = (sub_ds_size * sub_ds_count + (world - 1) * int(not drop_last)) // world
        # with the above parameters (if shuffle = False), the data idxs that rank 0 would generate are [0, 30), totally 30 elements
        # since batch size = 4, the batch number would be 8, with last batch only containing 2 elements
        # thus worker 0 of rank 0 would generate [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
        #      worker 1 of rank 0 would generate [16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29]
        sampler = AncSampler(ds, batch_size, world, rank, num_workers, shuffle=False, drop_last=drop_last, resume_step=resume_step)
        assert len(sampler) == (num_per_rank + batch_size - 1) // batch_size
        indices = []
        for item in sampler:
            cur_index, is_last_batch = item
            if not is_last_batch:
                # cur_index would be [] if resume_step % num_workers != 0
                assert len(cur_index) == batch_size or len(cur_index) == 0
            if is_last_batch:
                # each worker's last batch would set is_last_batch to True
                # so we have num_workers is_last_batch
                assert sampler.step >= len(sampler) - num_workers
            indices += cur_index
        if not shuffle and resume_step == 0:
            assert indices[:16] == [0, 1, 2, 3, 16, 17, 18, 19, 4, 5, 6, 7, 20, 21, 22, 23]
            assert indices[-6:] == [12, 13, 14, 15, 28, 29]
        return indices
    
    batch_size = 4
    from_beginning_indices = check_sampler_indices(shuffle=False, batch_size=batch_size, resume_step=0)
    check_sampler_indices(shuffle=True, batch_size=batch_size, resume_step=0)
    resume_step = 3
    resumed_indices = check_sampler_indices(shuffle=False, batch_size=batch_size, resume_step=resume_step)
    assert resumed_indices == from_beginning_indices[batch_size * resume_step:]

